"""
# The Python Data Valuation Library API

This is the API documentation for the Python Data Valuation Library (PyDVL).
Use the table of contents to access the documentation for each module.

The two main modules you will want to look at are [value][pydvl.value] and
[influence][pydvl.influence].
"""
__version__ = "0.8.0"
